from .embedder import MistralAIEmbedder
from .embedder import MistralAIEmbedderConfig
from .llm import MistralAILLM
from .llm import MistralAILLMConfig

__all__ = [
    'MistralAIEmbedder',
    'MistralAIEmbedderConfig',
    'MistralAILLMConfig',
    'MistralAILLM',
]



from mistralai import AssistantMessage
from mistralai import SystemMessage
from mistralai import ToolMessage
from mistralai import UserMessage

ChatCompletionMessageParams = (
    AssistantMessage | SystemMessage | UserMessage | ToolMessage
)



from __future__ import annotations

import os
from typing import Literal

from mistralai import Mistral
from pydantic import Field
from pydantic import SecretStr

from tinygent.core.datamodels.embedder import AbstractEmbedder
from tinygent.core.datamodels.embedder import AbstractEmbedderConfig
from tinygent.core.telemetry.decorators import tiny_trace
from tinygent.core.telemetry.utils import set_embedder_telemetry_attributes

_SUPPORTED_MODELS: dict[str, int] = {
    'mistral-embed': 1024,
    'codestral-embed': 1536,
}


class MistralAIEmbedderConfig(AbstractEmbedderConfig['MistralAIEmbedder']):
    type: Literal['mistralai'] = Field(default='mistralai', frozen=True)

    model: str = Field(default='mistral-embed')

    api_key: SecretStr | None = Field(
        default_factory=lambda: (
            SecretStr(os.environ['MISTRALAI_API_KEY'])
            if 'MISTRALAI_API_KEY' in os.environ
            else None
        ),
    )

    timeout: float = Field(default=60.0)

    def build(self) -> MistralAIEmbedder:
        return MistralAIEmbedder(
            model=self.model,
            api_key=self.api_key.get_secret_value() if self.api_key else None,
            timeout=self.timeout,
        )


class MistralAIEmbedder(AbstractEmbedder):
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        timeout: float = 60.0,
    ) -> None:
        if not api_key and not (api_key := os.getenv('MISTRALAI_API_KEY', None)):
            raise ValueError(
                'MistralAI API key must be provided either via config'
                "or 'MISTRALAI_API_KEY' env variable."
            )

        if model not in _SUPPORTED_MODELS:
            raise ValueError(
                f'Provided model name: {model} not in supported model names: {", ".join(_SUPPORTED_MODELS.keys())}'
            )

        self._client: Mistral | None = None
        self._model = model

        self.timeout = timeout
        self.api_key = api_key

    @property
    def model(self) -> str:
        return self._model

    @property
    def config(self) -> MistralAIEmbedderConfig:
        return MistralAIEmbedderConfig(
            model=self.model,
            api_key=SecretStr(self.api_key),
            timeout=self.timeout,
        )

    @property
    def embedding_dim(self) -> int:
        v = _SUPPORTED_MODELS.get(self.model)
        if not v:
            raise ValueError(
                f'Provided model name: {self.model} not in supported model names: {", ".join(_SUPPORTED_MODELS.keys())}'
            )
        return v

    def __get_client(self) -> Mistral:
        if self._client:
            return self._client

        self._client = Mistral(api_key=self.api_key, timeout_ms=int(self.timeout) * 1000)
        return self._client

    @tiny_trace('embed')
    def embed(self, query: str) -> list[float]:
        res = self.__get_client().embeddings.create(
            model=self.model,
            inputs=[query],
            timeout_ms=int(self.timeout) * 1000,
        )

        embedding = res.data[0].embedding
        if not embedding:
            raise ValueError(f'Error while creating embeddings for query {query}')

        set_embedder_telemetry_attributes(
            self.config,
            query,
            embedding_dim=self.embedding_dim,
            result_len=len(embedding),
        )
        return embedding

    @tiny_trace('embed_batch')
    def embed_batch(self, queries: list[str]) -> list[list[float]]:
        res = self.__get_client().embeddings.create(
            model=self.model,
            inputs=queries,
            timeout_ms=int(self.timeout) * 1000,
        )

        embeddings = []
        for q, r in zip(queries, res.data, strict=True):
            if (e := r.embedding) is None:
                raise ValueError(f'Error while creating embeddings for query {q}')
            embeddings.append(e)

        set_embedder_telemetry_attributes(
            self.config,
            queries,
            embedding_dim=self.embedding_dim,
            result_len=len(embeddings),
        )
        return embeddings

    @tiny_trace('aembed')
    async def aembed(self, query: str) -> list[float]:
        res = await self.__get_client().embeddings.create_async(
            model=self.model,
            inputs=[query],
            timeout_ms=int(self.timeout) * 1000,
        )

        embedding = res.data[0].embedding
        if not embedding:
            raise ValueError(f'Error while creating embeddings for query {query}')

        set_embedder_telemetry_attributes(
            self.config,
            query,
            embedding_dim=self.embedding_dim,
            result_len=len(embedding),
        )
        return embedding

    @tiny_trace('aembed_batch')
    async def aembed_batch(self, queries: list[str]) -> list[list[float]]:
        res = await self.__get_client().embeddings.create_async(
            model=self.model,
            inputs=queries,
            timeout_ms=int(self.timeout) * 1000,
        )

        embeddings = []
        for q, r in zip(queries, res.data, strict=True):
            if (e := r.embedding) is None:
                raise ValueError(f'Error while creating embeddings for query {q}')
            embeddings.append(e)

        set_embedder_telemetry_attributes(
            self.config,
            queries,
            embedding_dim=self.embedding_dim,
            result_len=len(embeddings),
        )
        return embeddings



def _register_mistralai() -> None:
    from tinygent.core.runtime.global_registry import GlobalRegistry

    from .embedder import MistralAIEmbedder
    from .embedder import MistralAIEmbedderConfig
    from .llm import MistralAILLM
    from .llm import MistralAILLMConfig

    registry = GlobalRegistry().get_registry()

    registry.register_llm('mistralai', MistralAILLMConfig, MistralAILLM)
    registry.register_embedder('mistralai', MistralAIEmbedderConfig, MistralAIEmbedder)


_register_mistralai()



from __future__ import annotations

from collections.abc import AsyncIterator
from functools import lru_cache
from io import StringIO
import json
import os
import textwrap
import typing
from typing import Iterable
from typing import Literal
from typing import override

from mistralai import Function
from mistralai import Mistral
from mistralai import Tool
from pydantic import Field
from pydantic import SecretStr
from transformers import AutoTokenizer
from transformers import PreTrainedTokenizerBase

from tiny_mistralai.utils import mistralai_chunk_to_tiny_chunks
from tiny_mistralai.utils import mistralai_family_to_tokenizer
from tiny_mistralai.utils import mistralai_result_to_tiny_result
from tiny_mistralai.utils import tiny_prompt_to_mistralai_params
from tinygent.core.datamodels.llm import AbstractLLM
from tinygent.core.datamodels.llm import AbstractLLMConfig
from tinygent.core.datamodels.messages import AllTinyMessages
from tinygent.core.telemetry.decorators import tiny_trace
from tinygent.core.telemetry.otel import set_tiny_attribute
from tinygent.core.telemetry.utils import set_llm_telemetry_attributes
from tinygent.core.types.io.llm_io_chunks import TinyLLMResultChunk
from tinygent.llms.utils import accumulate_llm_chunks
from tinygent.llms.utils import group_chunks_for_telemetry

if typing.TYPE_CHECKING:
    from tinygent.core.datamodels.llm import LLMStructuredT
    from tinygent.core.datamodels.tool import AbstractTool
    from tinygent.core.types.io.llm_io_input import TinyLLMInput
    from tinygent.core.types.io.llm_io_result import TinyLLMResult


class MistralAILLMConfig(AbstractLLMConfig['MistralAILLM']):
    type: Literal['mistralai'] = Field(default='mistralai', frozen=True)

    model: str = Field(default='mistral-medium-latest')

    api_key: SecretStr | None = Field(
        default_factory=lambda: (
            SecretStr(os.environ['MISTRALAI_API_KEY'])
            if 'MISTRALAI_API_KEY' in os.environ
            else None
        ),
    )

    safe_prompt: bool = Field(default=True)

    temperature: float = Field(default=0.6)

    timeout: float = Field(default=60.0)

    def build(self) -> MistralAILLM:
        return MistralAILLM(
            model=self.model,
            api_key=self.api_key.get_secret_value() if self.api_key else None,
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout=self.timeout,
        )


class MistralAILLM(AbstractLLM[MistralAILLMConfig]):
    def __init__(
        self,
        model: str = 'mistral-medium-latest',
        api_key: str | None = None,
        safe_prompt: bool = True,
        temperature: float = 0.6,
        timeout: float = 60.0,
    ) -> None:
        if not api_key and not (api_key := os.getenv('MISTRALAI_API_KEY', None)):
            raise ValueError(
                'MistralAI API key must be provided either via config'
                "or 'MISTRALAI_API_KEY' env variable."
            )

        self._client: Mistral | None = None

        self.model = model
        self.api_key = api_key
        self.safe_prompt = safe_prompt
        self.temperature = temperature
        self.timeout = timeout

    @property
    def config(self) -> MistralAILLMConfig:
        return MistralAILLMConfig(
            model=self.model,
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout=self.timeout,
            api_key=SecretStr(self.api_key),
        )

    @property
    def supports_tool_calls(self) -> bool:
        return True  # INFO: Not all models may support tool calls, but mistralai api error if not.

    def __get_client(self) -> Mistral:
        if self._client:
            return self._client

        self._client = Mistral(api_key=self.api_key, timeout_ms=int(self.timeout) * 1000)
        return self._client

    @override
    def _tool_convertor(self, tool: AbstractTool) -> Tool:
        info = tool.info
        schema = info.input_schema

        def map_type(py_type: type) -> str:
            mapping = {
                str: 'string',
                int: 'integer',
                float: 'number',
                bool: 'boolean',
                list: 'array',
                dict: 'object',
            }
            return mapping.get(py_type, 'string')  # default fallback

        properties = {}

        if schema:
            for name, field in schema.model_fields.items():
                field_type = (
                    field.annotation
                    if isinstance(field.annotation, type)
                    else type(field.annotation)
                )

                prop = {'type': map_type(field_type)}
                if field.description:
                    prop['description'] = field.description

                properties[name] = prop

        return Tool(
            type='function',
            function=Function(
                name=info.name,
                description=info.description,
                parameters={
                    'type': 'object',
                    'properties': properties,
                    'required': info.required_fields,
                },
            ),
        )

    @staticmethod
    @lru_cache(maxsize=None)
    def _get_encoding(model: str) -> PreTrainedTokenizerBase:
        return AutoTokenizer.from_pretrained(
            mistralai_family_to_tokenizer(model), use_fast=True
        )

    @staticmethod
    @lru_cache(maxsize=100_000)
    def _count_tokens(model: str, text: str) -> int:
        return len(MistralAILLM._get_encoding(model).encode(text))

    @tiny_trace('generate_text')
    def generate_text(
        self,
        llm_input: TinyLLMInput,
    ) -> TinyLLMResult:
        messages = tiny_prompt_to_mistralai_params(llm_input)

        res = self.__get_client().chat.complete(
            model=self.model,
            messages=messages,
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout_ms=int(self.timeout * 1000),
        )

        tiny_res = mistralai_result_to_tiny_result(res)
        set_llm_telemetry_attributes(
            self.config, llm_input.messages, result=tiny_res.to_string()
        )
        return tiny_res

    @tiny_trace('agenerate_text')
    async def agenerate_text(
        self,
        llm_input: TinyLLMInput,
    ) -> TinyLLMResult:
        messages = tiny_prompt_to_mistralai_params(llm_input)

        res = await self.__get_client().chat.complete_async(
            model=self.model,
            messages=messages,
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout_ms=int(self.timeout * 1000),
        )

        tiny_res = mistralai_result_to_tiny_result(res)
        set_llm_telemetry_attributes(
            self.config, llm_input.messages, result=tiny_res.to_string()
        )
        return tiny_res

    @tiny_trace('stream_text')
    async def stream_text(
        self, llm_input: TinyLLMInput
    ) -> AsyncIterator[TinyLLMResultChunk]:
        messages = tiny_prompt_to_mistralai_params(llm_input)
        set_llm_telemetry_attributes(self.config, llm_input.messages)

        res = await self.__get_client().chat.stream_async(
            model=self.model,
            messages=messages,
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout_ms=int(self.timeout * 1000),
        )

        async def raw_chunks() -> AsyncIterator[TinyLLMResultChunk]:
            async for chunk in res:
                for tiny_chunk in mistralai_chunk_to_tiny_chunks(chunk.data):
                    yield tiny_chunk

        accumulated_chunks: list[TinyLLMResultChunk] = []
        try:
            async for acc_chunk in accumulate_llm_chunks(raw_chunks()):
                accumulated_chunks.append(acc_chunk)
                yield acc_chunk
        finally:
            set_tiny_attribute(
                'result',
                group_chunks_for_telemetry(accumulated_chunks),
            )

    @tiny_trace('generate_structured')
    def generate_structured(
        self, llm_input: TinyLLMInput, output_schema: type[LLMStructuredT]
    ) -> LLMStructuredT:
        messages = tiny_prompt_to_mistralai_params(llm_input)

        res = self.__get_client().chat.parse(
            model=self.model,
            messages=messages,
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout_ms=int(self.timeout * 1000),
            response_format=output_schema,
        )

        if not res.choices or not (message := res.choices[0].message):
            raise ValueError('No message in MistralAI response.')

        parsed = output_schema.model_validate(json.loads(str(message.content) or '{}'))
        set_llm_telemetry_attributes(
            self.config,
            llm_input.messages,
            result=str(parsed),
            output_schema=output_schema,
        )
        return parsed

    @tiny_trace('agenerate_structured')
    async def agenerate_structured(
        self, llm_input: TinyLLMInput, output_schema: type[LLMStructuredT]
    ) -> LLMStructuredT:
        messages = tiny_prompt_to_mistralai_params(llm_input)

        res = await self.__get_client().chat.parse_async(
            model=self.model,
            messages=messages,
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout_ms=int(self.timeout * 1000),
            response_format=output_schema,
        )

        if not res.choices or not (message := res.choices[0].message):
            raise ValueError('No message in MistralAI response.')

        parsed = output_schema.model_validate(json.loads(str(message.content) or '{}'))
        set_llm_telemetry_attributes(
            self.config,
            llm_input.messages,
            result=str(parsed),
            output_schema=output_schema,
        )
        return parsed

    @tiny_trace('generate_with_tools')
    def generate_with_tools(
        self, llm_input: TinyLLMInput, tools: list[AbstractTool]
    ) -> TinyLLMResult:
        functions = [self._tool_convertor(tool) for tool in tools]
        messages = tiny_prompt_to_mistralai_params(llm_input)

        res = self.__get_client().chat.complete(
            model=self.model,
            messages=messages,
            tools=functions,
            tool_choice='auto',
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout_ms=int(self.timeout * 1000),
        )

        tiny_res = mistralai_result_to_tiny_result(res)
        set_llm_telemetry_attributes(
            self.config, llm_input.messages, result=tiny_res.to_string(), tools=tools
        )
        return tiny_res

    @tiny_trace('agenerate_with_tools')
    async def agenerate_with_tools(
        self, llm_input: TinyLLMInput, tools: list[AbstractTool]
    ) -> TinyLLMResult:
        functions = [self._tool_convertor(tool) for tool in tools]
        messages = tiny_prompt_to_mistralai_params(llm_input)

        res = await self.__get_client().chat.complete_async(
            model=self.model,
            messages=messages,
            tools=functions,
            tool_choice='auto',
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout_ms=int(self.timeout * 1000),
        )

        tiny_res = mistralai_result_to_tiny_result(res)
        set_llm_telemetry_attributes(
            self.config, llm_input.messages, result=tiny_res.to_string(), tools=tools
        )
        return tiny_res

    @tiny_trace('stream_with_tools')
    async def stream_with_tools(
        self, llm_input: TinyLLMInput, tools: list[AbstractTool]
    ) -> AsyncIterator[TinyLLMResultChunk]:
        functions = [self._tool_convertor(tool) for tool in tools]
        messages = tiny_prompt_to_mistralai_params(llm_input)
        set_llm_telemetry_attributes(self.config, llm_input.messages, tools=tools)

        res = await self.__get_client().chat.stream_async(
            model=self.model,
            messages=messages,
            tools=functions,
            tool_choice='auto',
            safe_prompt=self.safe_prompt,
            temperature=self.temperature,
            timeout_ms=int(self.timeout * 1000),
        )

        async def raw_chunks() -> AsyncIterator[TinyLLMResultChunk]:
            async for chunk in res:
                for tiny_chunk in mistralai_chunk_to_tiny_chunks(chunk.data):
                    yield tiny_chunk

        accumulated_chunks: list[TinyLLMResultChunk] = []
        try:
            async for acc_chunk in accumulate_llm_chunks(raw_chunks()):
                accumulated_chunks.append(acc_chunk)
                yield acc_chunk
        finally:
            set_tiny_attribute(
                'result',
                group_chunks_for_telemetry(accumulated_chunks),
            )

    def count_tokens_in_messages(self, messages: Iterable[AllTinyMessages]) -> int:
        set_llm_telemetry_attributes(self.config, messages)

        number_of_tokens = sum(
            [MistralAILLM._count_tokens(self.model, m.tiny_str) for m in messages]
        )

        set_tiny_attribute('number_of_tokens', number_of_tokens)
        return number_of_tokens

    def __str__(self) -> str:
        buf = StringIO()

        buf.write('MistralAI LLM Summary:\n')
        buf.write(textwrap.indent(f'Model: {self.model}\n', '\t'))
        buf.write(textwrap.indent(f'Safe Prompt: {self.safe_prompt}\n', '\t'))
        buf.write(textwrap.indent(f'Temperature: {self.temperature}\n', '\t'))
        buf.write(textwrap.indent(f'Timeout: {self.timeout}\n', '\t'))

        return buf.getvalue()



import typing
from typing import cast

from langchain_core.messages import AIMessage
from langchain_core.outputs import ChatGeneration
from langchain_core.outputs import Generation
from mistralai import UNSET
from mistralai import AssistantMessage
from mistralai import ChatCompletionResponse
from mistralai import CompletionChunk
from mistralai import Content
from mistralai import FunctionCall
from mistralai import SystemMessage
from mistralai import TextChunk
from mistralai import ToolCall
from mistralai import ToolMessage
from mistralai import UserMessage

from tiny_mistralai.types import ChatCompletionMessageParams
from tinygent.core.datamodels.messages import TinyChatMessage
from tinygent.core.datamodels.messages import TinyChatMessageChunk
from tinygent.core.datamodels.messages import TinyHumanMessage
from tinygent.core.datamodels.messages import TinyPlanMessage
from tinygent.core.datamodels.messages import TinyReasoningMessage
from tinygent.core.datamodels.messages import TinySystemMessage
from tinygent.core.datamodels.messages import TinyToolCall
from tinygent.core.datamodels.messages import TinyToolResult
from tinygent.core.types.io.llm_io_chunks import TinyLLMResultChunk
from tinygent.core.types.io.llm_io_chunks import TinyToolCallChunk
from tinygent.core.types.io.llm_io_result import TinyLLMResult

if typing.TYPE_CHECKING:
    from tinygent.core.types.io.llm_io_input import TinyLLMInput


def _normalize_content(content: Content) -> str:
    """Normalize Mistral AI Content to a string."""
    if isinstance(content, str):
        return content
    elif isinstance(content, list):
        return ''.join(chunk.text for chunk in content if isinstance(chunk, TextChunk))
    else:
        return ''


def tiny_prompt_to_mistralai_params(
    prompt: 'TinyLLMInput',
) -> list[ChatCompletionMessageParams]:
    params: list[ChatCompletionMessageParams] = []

    for msg in prompt.messages:
        if isinstance(msg, TinyHumanMessage):
            params.append(UserMessage(role='user', content=str(msg.content)))

        elif isinstance(msg, TinySystemMessage):
            params.append(SystemMessage(role='system', content=str(msg.content)))

        elif isinstance(msg, TinyChatMessage):
            params.append(AssistantMessage(role='assistant', content=msg.content))

        elif isinstance(msg, TinyPlanMessage):
            params.append(
                AssistantMessage(
                    role='assistant', content=f'<PLAN>\n{msg.content}\n</PLAN>'
                )
            )

        elif isinstance(msg, TinyReasoningMessage):
            params.append(
                AssistantMessage(
                    role='assistant',
                    content=f'<REASONING>\n{msg.content}\n</REASONING>',
                )
            )

        elif isinstance(msg, TinyToolCall):
            params.append(
                AssistantMessage(
                    role='assistant',
                    content=None,
                    tool_calls=[
                        ToolCall(
                            id=msg.call_id or 'tool_call_1',
                            type='function',
                            function=FunctionCall(
                                name=msg.tool_name, arguments=str(msg.arguments)
                            ),
                        )
                    ],
                )
            )

        elif isinstance(msg, TinyToolResult):
            params.append(
                ToolMessage(
                    role='tool',
                    content=msg.content,
                    tool_call_id=msg.call_id,
                )
            )

        else:
            raise TypeError(f'Unsupported TinyMessage type: {type(msg)}')

    return params


def mistralai_result_to_tiny_result(resp: ChatCompletionResponse) -> TinyLLMResult:
    """Convert a Mistral AI ChatCompletionResponse to a TinyLLMResult."""
    generations: list[list[Generation]] = []

    for choice in resp.choices:
        msg = choice.message
        text = _normalize_content(msg.content) if msg.content else ''

        additional_kwargs = {}
        if getattr(msg, 'tool_calls', None):
            tool_calls = []
            for tc in msg.tool_calls or []:
                tool_calls.append(
                    {
                        'id': tc.id,
                        'type': 'function',
                        'function': {
                            'name': tc.function.name,
                            'arguments': tc.function.arguments,
                        },
                    }
                )
            additional_kwargs['tool_calls'] = tool_calls

        ai_msg = AIMessage(content=text, additional_kwargs=additional_kwargs)
        generations.append([ChatGeneration(message=ai_msg, text=text)])

    llm_output = {
        'id': resp.id,
        'model': resp.model,
        'created': resp.created,
        'usage': resp.usage.model_dump() if resp.usage else None,
        'finish_reasons': [c.finish_reason for c in resp.choices],
    }

    return TinyLLMResult(generations=generations, llm_output=llm_output)


def mistralai_chunk_to_tiny_chunk(chunk: CompletionChunk) -> TinyLLMResultChunk:
    """Convert a Mistral AI CompletionChunk to a TinyLLMResultChunk."""
    choice = chunk.choices[0]
    delta = choice.delta

    # tool call chunk
    if (tcs := delta.tool_calls) is not UNSET and tcs is not None:
        for tc in cast(list[ToolCall], tcs):
            if tc.function:
                return TinyLLMResultChunk(
                    type='tool_call',
                    tool_call=TinyToolCallChunk(
                        tool_name=tc.function.name or '',
                        arguments=str(tc.function.arguments) or '',
                        call_id=tc.id,
                        index=tc.index or 0,
                        metadata={'raw': tc.model_dump()},
                    ),
                )

    # text chunk
    if delta.content:
        return TinyLLMResultChunk(
            type='message',
            message=TinyChatMessageChunk(
                content=_normalize_content(delta.content),
                metadata={'raw': delta.model_dump()},
            ),
            metadata={'finish_reason': choice.finish_reason},
        )

    # end of stream
    if choice.finish_reason is not None:
        return TinyLLMResultChunk(
            type='end',
            metadata={'finish_reason': choice.finish_reason},
        )

    return TinyLLMResultChunk(type='none', metadata={'raw': delta.model_dump()})


def mistralai_chunk_to_tiny_chunks(chunk: CompletionChunk) -> list[TinyLLMResultChunk]:
    """Convert a Mistral AI CompletionChunk to a list of TinyLLMResultChunks."""
    choice = chunk.choices[0]
    delta = choice.delta

    results: list[TinyLLMResultChunk] = []

    if (tcs := delta.tool_calls) is not UNSET and tcs is not None:
        for tc in cast(list[ToolCall], tcs):
            if tc.function:
                results.append(
                    TinyLLMResultChunk(
                        type='tool_call',
                        tool_call=TinyToolCallChunk(
                            tool_name=tc.function.name or '',
                            arguments=str(tc.function.arguments) or '',
                            call_id=tc.id,
                            index=tc.index or 0,
                            metadata={'raw': tc.model_dump()},
                        ),
                    )
                )

    if delta.content:
        results.append(
            TinyLLMResultChunk(
                type='message',
                message=TinyChatMessageChunk(
                    content=_normalize_content(delta.content),
                    metadata={'raw': delta.model_dump()},
                ),
                metadata={'finish_reason': choice.finish_reason},
            )
        )

    if choice.finish_reason is not None:
        results.append(
            TinyLLMResultChunk(
                type='end',
                metadata={'finish_reason': choice.finish_reason},
            )
        )

    if not results:
        results.append(
            TinyLLMResultChunk(type='none', metadata={'raw': delta.model_dump()})
        )

    return results


def mistralai_family_to_tokenizer(model: str) -> str:
    m = model.lower()

    if m.startswith('codestral'):
        return 'mistralai/Codestral-22B-v0.1'

    if m.startswith('ministral'):
        return 'mistralai/Ministral-8B-Instruct-2410'

    if m.startswith('pixtral'):
        return 'mistralai/Pixtral-12B-2409'

    if m.startswith('devstral'):
        return 'mistralai/Mistral-7B-Instruct-v0.2'

    if m.startswith('magistral'):
        if 'small' in m:
            return 'mistralai/Mistral-7B-Instruct-v0.2'
        return 'mistralai/Mixtral-8x7B-Instruct-v0.1'

    if m.startswith('mistral'):
        if 'large' in m:
            return 'mistralai/Mixtral-8x22B-Instruct-v0.1'
        if 'medium' in m:
            return 'mistralai/Mixtral-8x7B-Instruct-v0.1'
        return 'mistralai/Mistral-7B-Instruct-v0.2'

    return 'mistralai/Mistral-7B-Instruct-v0.2'
